import 'dart:async';

import 'package:dartz/dartz.dart';
import 'package:flutter_gb_stack_base/flutter_gb_stack_base.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:get_it/get_it.dart';
import 'package:logger/logger.dart';
import 'package:rxdart/rxdart.dart';

import '../../../flutter_gb_authentication_basic.dart';

part 'authentication_basic_bloc.freezed.dart';
part 'authentication_basic_bloc_event.dart';
part 'authentication_basic_bloc_state.dart';

class AuthenticationBasicBloc extends ListenableBloc<AuthenticationBasicEvent, AuthenticationBasicState> {
  final AuthenticationService authenticationService;
  final BasicAuthenticationConfig config;

  bool initialAuthChecked = false;

  late Stream<AuthenticationBasicEvent> eventsReplay;

  late StreamSubscription _refreshEventsSubscription;

  AuthenticationBasicBloc({
    required this.authenticationService,
    required this.config,
    Logger? logger,
  }) : super(
          AuthenticationBasicState.initial(),
          logger: logger ?? getAuthenticationBasicLogger(config),
        ) {
    // register so we can get this instance later on on other libraries or other part of the application
    if (!GetIt.instance.isRegistered<AuthenticationBasicBloc>()) {
      GetIt.instance.registerSingleton<AuthenticationBasicBloc>(this);
    }

    eventsReplay = ReplaySubject<AuthenticationBasicEvent>(
      maxSize: 10,
    )..addStream(events());

    _refreshEventsSubscription = authRefreshEvents.listen((event) {
      add(
        event.fold(
          (l) => l,
          (r) => r,
        ),
      );
    });

    on<GetSession>((event, emit) async {
      final authSession = await authenticationService.getAuthenticatedSession();
      authSession.fold(
        (failure) {
          emit(
            state.copyWith(
              authenticatingFailure: failure,
              authChecked: true,
            ),
          );
          if (failure is NoParsableUserFailure) {
            add(
              AuthenticationBasicEvent.logout(
                sessionToken: failure.partialData.token,
              ),
            );
          } else {
            add(
              const AuthenticationBasicEvent.logout(
                sessionToken: null,
              ),
            );
          }
        },
        (authData) {
          final isAuthenticated = authData != null;
          emit(
            state.copyWith(
              authChecked: true,
              authenticationData: authData,
            ),
          );
          if (isAuthenticated) {
            add(
              AuthenticationBasicEvent.authChanged(
                authenticated: true,
                data: authData,
              ),
            );
          }
          if (isAuthenticated && !initialAuthChecked && !config.initialAuthCheckedAfterUserFetch) {
            initialAuthChecked = true;
            add(
              AuthenticationBasicEvent.initialAuthChecked(authData),
            );
          } else if (!isAuthenticated) {
            initialAuthChecked = true;
            add(
              AuthenticationBasicEvent.initialAuthChecked(authData),
            );
          }
        },
      );
    });

    on<LoginWithCredentials>((event, emit) async {
      emit(
        state.copyWith(
          authenticating: true,
        ),
      );

      final response = await authenticationService.loginWithCredentials(event.email, event.password);
      // await Future.delayed(Duration(milliseconds: 800));

      add(
        AuthenticationBasicEvent.loginWithCredentialsRes(response: response),
      );
    });

    on<AuthChanged>((event, emit) async {
      emit(
        state.copyWith(
          authenticationData: event.data,
        ),
      );

      // fetch user information if Endpoint provided
      if (event.authenticated && config.fetchUserInformationAPIendpoint != null) {
        add(
          FetchAuthUserData(authData: event.data!),
        );
      }
    });

    on<LoginWithCredentialsRes>((event, emit) async {
      event.response.fold(
        (failure) {
          emit(
            state.copyWith(
              authenticatingFailure: failure,
              authenticating: false,
            ),
          );
        },
        (data) {
          emit(
            state.copyWith(
              authenticating: false,
              authenticationData: data,
            ),
          );
          add(
            AuthenticationBasicEvent.authChanged(authenticated: true, data: data),
          );
        },
      );
    });

    on<Logout>((event, emit) async {
      // wait a little bit before removing session data, so if someone hooks on the event they have a chance to
      // perform some action with the default auth headers
      await Future.delayed(const Duration(milliseconds: 250));
      final res = await authenticationService.logout();

      emit(res.fold(
        (l) {
          // state.copyWith(
          // );
          return state;
        },
        (r) {
          return state.copyWith(
            authenticationData: null,
          );
        },
      ));

      if (isClosed) return;

      add(
        const AuthenticationBasicEvent.authChanged(
          authenticated: false,
          data: null,
        ),
      );
    });

    on<SignupWithCredentials>((event, emit) async {
      emit(
        state.copyWith(
          signingUp: true,
        ),
      );

      final res = await authenticationService.signupWithCredentials(event.payload);
      add(
        SignupWithCredentialsResult(
          result: res,
          payload: event.payload,
        ),
      );
    });

    on<SignupWithCredentialsResult>((event, emit) async {
      event.result.fold(
        (failure) {
          emit(
            state.copyWith(
              signingUp: false,
              signingUpFailure: failure,
            ),
          );
        },
        (authData) async {
          emit(
            state.copyWith(
              signingUp: false,
              signingUpFailure: null,
            ),
          );

          /// try to perform login if defined on configuration
          if (config.authenticateOnSignup) {
            /// first check if we have a valid authData returned by the signup
            if (authData != null) {
              /// if so then emit authentication change
              add(
                AutoSignupAuthResult(
                  result: right<Failure, AuthenticationData>(
                    authData,
                  ),
                ),
              );
              // add(
              //   AuthChanged(authenticated: true, data: authData),
              // );
            }
            // if no auth data returned from signup call, let's try one more time with user custom defined callback
            else if (config.customAutoSignupAuthClbk != null) {
              emit(
                state.copyWith(
                  authenticating: true,
                ),
              );
              final credentials = await config.customAutoSignupAuthClbk!(event.payload, authData);
              final loginRes = await authenticationService.loginWithCredentials(
                credentials.usernameOrEmail,
                credentials.password,
              );

              add(
                AutoSignupAuthResult(
                  result: loginRes,
                ),
              );
            } else {
              /// notify about the failure
              add(
                AutoSignupAuthResult(
                  result: left<Failure, AuthenticationData>(
                    UnknownFailure(
                      message: "Unable to perform Authentication, please try again later.",
                    ),
                  ),
                ),
              );
            }
          }
        },
      );
    });

    on<AutoSignupAuthResult>((event, emit) {
      emit(
        state.copyWith(
          authenticating: false,
        ),
      );
      event.result.fold(
        (l) => null,
        (authData) {
          add(
            AuthChanged(authenticated: true, data: authData),
          );
        },
      );
    });

    on<FetchAuthUserData>((event, emit) async {
      if (!state.authenticated) {
        logger?.w("User is not authenticated, so user fetch is not going to be fetched. Will throw a failure");
      }
      emit(
        state.copyWith(
          fetchingAuthUserData: true,
        ),
      );
      final res = !state.authenticated
          ? left<Failure, User>(
              NoAuthenticatedFailure(),
            )
          : await authenticationService.fetchAuthUserData();
      if (!initialAuthChecked && config.initialAuthCheckedAfterUserFetch) {
        initialAuthChecked = true;
        res.fold(
          (l) {
            add(
              AuthenticationBasicEvent.initialAuthChecked(
                event.authData,
              ),
            );
          },
          (user) {
            add(
              AuthenticationBasicEvent.initialAuthChecked(
                (event.authData as AuthenticationDataModel).copyWith(
                  user: user,
                ),
              ),
            );
          },
        );
      }
      await Future.delayed(const Duration(milliseconds: 100));
      add(
        FetchAuthUserDataResult(result: res),
      );
    });

    on<FetchAuthUserDataResult>((event, emit) {
      emit(
        event.result.fold(
          (failure) {
            if (failure is SessionExpiredFailure) {
              add(
                SessionExpired(
                  authData: state.authenticationData!,
                ),
              );
            }
            return state.copyWith(
              fetchingAuthUserData: false,
            );
          },
          (user) {
            return state.copyWith(
              fetchingAuthUserData: false,
              authenticationData: (state.authenticationData as AuthenticationDataModel?)?.copyWith(
                user: user,
              ),
            );
          },
        ),
      );
    });

    on<InitialAuthChecked>((event, emit) {
      if (event.data != null) {
        // let's check if session have expired
        // and if we would be able to refresh it if not notify a session expiration
        if (event.data!.expired && config.refreshTokenAPIendpoint == null) {
          if (isClosed) return;
          add(
            SessionExpired(authData: event.data!),
          );
        }
      }

      // fetch user info if authenticated
      // if (event.data != null) {
      //   add(
      //     FetchAuthUserData(authData: event.data!),
      //   );
      // }
    });

    on<SessionExpired>((event, emit) {
      if (config.autoLogoutOnSessionExpired) {
        logout();
      }
    });

    on<RefreshTokenFailure>((event, emit) {
      add(
        SessionExpired(
          authData: state.authenticationData ??
              AuthenticationDataModel.fromToken(
                event.token,
              ),
        ),
      );
    });

    on<RefreshTokenSuccess>((event, emit) async {
      final result = await authenticationService.getAuthenticatedSession();

      result.fold(
        (l) {},
        (data) {
          emit(
            state.copyWith(
              authenticationData: data,
            ),
          );
        },
      );
    });

    on<DeleteAccount>((event, emit) async {
      emit(
        state.copyWith(
          deletingAccount: true,
        ),
      );

      final res = await authenticationService.deleteAccount(state.authenticationData?.id ?? '');

      add(
        DeleteAccountResult(
          result: res,
        ),
      );
    });

    on<DeleteAccountResult>((event, emit) {
      event.result.fold(
        (failure) {
          emit(
            state.copyWith(
              deletingAccount: false,
              deleteAccountFailure: failure,
            ),
          );
        },
        (r) {
          emit(
            state.copyWith(
              deletingAccount: false,
              deleteAccountFailure: null,
            ),
          );
          if (config.logoutAfterDelete) {
            logout();
          }
        },
      );
    });
  }

  @override
  @mustCallSuper
  Future<void> close() async {
    // actionsReplay;
    _refreshEventsSubscription.cancel();
    return super.close();
  }

  getSession() {
    if (isClosed) return;
    add(
      const GetSession(),
    );
  }

  loginWithCredentials(String email, String password) {
    if (isClosed) return;
    add(
      LoginWithCredentials(email: email, password: password),
    );
  }

  signupWithCredentials(AuthSignUpPayload payload) {
    if (isClosed) return;
    add(
      SignupWithCredentials(payload: payload),
    );
  }

  logout() {
    if (isClosed) return;
    if (state.authenticated) {
      add(
        Logout(sessionToken: state.authenticationData!.token),
      );
    }
  }

  fetchUser() {
    if (isClosed) return;
    if (state.authenticationData == null) {
      return add(
        FetchAuthUserDataResult(
          result: left(
            NoAuthenticatedFailure(),
          ),
        ),
      );
    }
    add(
      FetchAuthUserData(
        authData: state.authenticationData!,
      ),
    );
  }

  deleteAccount() {
    if (isClosed) return;
    add(
      const DeleteAccount(),
    );
  }
}
