import 'dart:async';

import 'package:corsac_jwt/corsac_jwt.dart';
import 'package:flutter_gb_stack_base/flutter_gb_stack_base.dart';

import '../flutter_gb_authentication_basic.dart';

class BasicAuthenticationConfig {
  /// Provides a flag for printing useful debugging data
  /// Defaults to [false]
  final bool debug;

  ///
  /// Provides an endpoint to be used when trying to login with credentials
  final Uri Function() loginCredentialsAPIendpoint;

  /// Provides an endpoint to be used when trying to signup
  final Uri Function() signupCredentialsAPIendpoint;

  /// Defines a function to the URL in which a new token can be requested
  /// Defining this property activates the 'refreshToken' mechanism
  final Uri Function(String refreshToken, JWT token)? refreshTokenAPIendpoint;

  ///
  /// Provides a way to transform Login response by providing RAW body
  ///
  final FutureOr<AuthenticationData> Function(String response)? customLoginResponseParser;

  ///
  /// Provides a custom user mappers that will be using when decoding a user from user data API and from
  /// auth session
  final FutureOr<User>? Function(Map<String, dynamic> userMap)? customUserMapper;

  ///
  /// Provides a custom function that allows you to customize the way your request's body is sent to the API
  /// Example
  /// function(email, password) {
  ///   return {
  ///     'custom-email-key': email,
  ///     'custom-password-key': password,
  ///   }
  /// }
  final FutureOr<Map<String, String>> Function(String email, String password)? customLoginRequestMapper;

  ///
  /// Provides a way to transform Signup with Credentials response by providing access to RAW body
  ///
  final FutureOr<AuthenticationData>? Function(String body)? customSignupCredentialsResponseParser;

  ///
  /// Specifies if the library should automatically authenticate the user after a successful signup
  /// This will be performed just if the system successfully extracts an [AuthenticationData] from the response
  /// or if user have defined a [customAutoSignupAuthClbk] callback
  /// if this is set t true and the library fails to authenticate it will notify the result on the bloc bia [AutoSignupAuthResult]
  /// Defaults to [true]
  final bool authenticateOnSignup;

  ///
  /// Provides a callback to be used when trying to signIn after a successful signup
  /// This is useful when no [AuthenticationData] was returned from the API call or may it's missing data
  /// It provides access to the submitted [AuthSignUpPayload] and the returned [AuthenticationData] if any.
  /// Should return a
  final FutureOr<AuthCredentialsProvider> Function(
    AuthSignUpPayload payload,
    AuthenticationData? authData,
  )? customAutoSignupAuthClbk;

  ///
  /// Provides an endpoint to be used when trying to fetch with credentials
  final Uri Function(AuthenticationData data)? fetchUserInformationAPIendpoint;

  ///
  /// Provides a way to transform fetchUserInformationAPIendpoint response to user map, then [customUserMapper] will be
  /// used to the
  final FutureOr<Map<String, dynamic>> Function(Map<String, dynamic> data)? customFetchUserInformationResponseMapper;

  ///
  /// Dispatch [initialAuthChecked] event just after we have an initial user fetch data
  /// Defaults to [true]
  final bool initialAuthCheckedAfterUserFetch;

  ///
  /// Dispatch [Logout] event if session expired error returned from fetchUserData API
  /// Defaults to [true]
  final bool autoLogoutOnSessionExpired;

  /// Defines a custom function to parse the response of your refresh token API, you must return an object
  /// with "refreshToken" and "authToken", so it can replace your current session one
  /// Example
  /// {
  ///   "auth-token": "YourParsedAuthToken",
  ///   "auth-refresh-token": "YourParsedRefreshToken"
  /// }
  final FutureOr<Map<String, String>> Function(String body)? customRefreshTokenResponseParser;

  /// Defines the 'http' (POST, PUT, etc) method to be used when requesting a new token to the API
  final String? refreshTokenMethod;

  /// Defines the JWT age thresshold in which the refresh token logic will trigger
  ///
  /// By default `1 day` is used
  final Duration? maxAge;

  /// This Callback is called whenever a refresh token have been successfully retrieved
  /// It provides access to the retrieved tokens
  final void Function(Map<String, String> tokens)? onRefreshToken;

  /// Provide a function to be used in order to use/pass the refresh token to the
  /// API endpoint
  final FutureOr<Map<String, String>> Function(String refreshToken, String authToken)?
      customRefreshTokenRequestBodyMapper;

  /// Define a custom period to wait for the refresh token API, after defined period if no response
  /// the API call waiting for the token will throw
  ///
  /// Defaults to 15 seconds
  ///
  final Duration? refreshTokenTimeout;

  /// This Callback is called whenever a refresh token fail to be retrieved
  /// It provides access to the refresh token used & the exception raised
  final void Function(String token, Object exception)? onRefreshTokenFailure;

  ///
  /// Provides an endpoint to be used when request an account delete
  final Uri Function(String userId) deleteAccountApiEndpoint;

  ///
  /// Provides a way to customize the way the response is parsed
  final User? Function(Map<String, dynamic> response)? deleteAccountCustomResponseParser;

  ///
  /// Determine if user should be logged out after account deletion
  /// Defaults to [true]
  final bool logoutAfterDelete;

  BasicAuthenticationConfig({
    required this.loginCredentialsAPIendpoint,
    required this.signupCredentialsAPIendpoint,
    this.customLoginResponseParser,
    this.customUserMapper,
    this.refreshTokenAPIendpoint,
    this.debug = false,
    this.customLoginRequestMapper,
    this.customSignupCredentialsResponseParser,
    this.authenticateOnSignup = true,
    this.customAutoSignupAuthClbk,
    this.fetchUserInformationAPIendpoint,
    this.customFetchUserInformationResponseMapper,
    this.initialAuthCheckedAfterUserFetch = true,
    this.autoLogoutOnSessionExpired = true,
    this.maxAge,
    this.customRefreshTokenResponseParser,
    this.onRefreshToken,
    this.refreshTokenMethod,
    this.customRefreshTokenRequestBodyMapper,
    this.refreshTokenTimeout,
    this.onRefreshTokenFailure,
    required this.deleteAccountApiEndpoint,
    this.deleteAccountCustomResponseParser,
    this.logoutAfterDelete = true,
  });

  BasicAuthenticationConfig copyWith({
    Uri Function()? loginCredentialsAPIendpoint,
    Uri Function()? signupCredentialsAPIendpoint,
    FutureOr<AuthenticationData> Function(String response)? customLoginResponseParser,
    FutureOr<User>? Function(Map<String, dynamic> userMap)? customUserMapper,
    Uri Function(String refreshToken, JWT token)? refreshTokenAPIendpoint,
    bool? debug,
    FutureOr<Map<String, String>> Function(String email, String password)? customLoginRequestMapper,
    FutureOr<AuthenticationData>? Function(String response)? customSignupCredentialsResponseParser,
    bool? authenticateOnSignup,
    AuthCredentialsProvider Function(AuthSignUpPayload payload, AuthenticationData? authData)? customAutoSignupAuthClbk,
    Uri Function(AuthenticationData data)? fetchUserInformationAPIendpoint,
    FutureOr<Map<String, dynamic>> Function(Map<String, dynamic> data)? customFetchUserInformationResponseMapper,
    bool? initialAuthCheckedAfterUserFetch,
    bool? autoLogoutOnSessionExpired,
    Duration? maxAge,
    FutureOr<Map<String, String>> Function(String body)? customRefreshTokenResponseParser,
    void Function(Map<String, String> tokens)? onRefreshToken,
    String? refreshTokenMethod,
    FutureOr<Map<String, String>> Function(String refreshToken, String authToken)? customRefreshTokenRequestBodyMapper,
    Duration? refreshTokenTimeout,
    void Function(String token, Object exception)? onRefreshTokenFailure,
    Uri Function(String userId)? deleteAccountApiEndpoint,
    User? Function(Map<String, dynamic> response)? deleteAccountCustomResponseParser,
    bool? logoutAfterDelete,
  }) {
    return BasicAuthenticationConfig(
      loginCredentialsAPIendpoint: loginCredentialsAPIendpoint ?? this.loginCredentialsAPIendpoint,
      signupCredentialsAPIendpoint: signupCredentialsAPIendpoint ?? this.signupCredentialsAPIendpoint,
      customLoginResponseParser: customLoginResponseParser ?? this.customLoginResponseParser,
      customUserMapper: customUserMapper ?? this.customUserMapper,
      refreshTokenAPIendpoint: refreshTokenAPIendpoint ?? this.refreshTokenAPIendpoint,
      debug: debug ?? this.debug,
      customLoginRequestMapper: customLoginRequestMapper ?? this.customLoginRequestMapper,
      customSignupCredentialsResponseParser:
          customSignupCredentialsResponseParser ?? this.customSignupCredentialsResponseParser,
      authenticateOnSignup: authenticateOnSignup ?? this.authenticateOnSignup,
      customAutoSignupAuthClbk: customAutoSignupAuthClbk ?? this.customAutoSignupAuthClbk,
      fetchUserInformationAPIendpoint: fetchUserInformationAPIendpoint ?? this.fetchUserInformationAPIendpoint,
      customFetchUserInformationResponseMapper:
          customFetchUserInformationResponseMapper ?? this.customFetchUserInformationResponseMapper,
      initialAuthCheckedAfterUserFetch: initialAuthCheckedAfterUserFetch ?? this.initialAuthCheckedAfterUserFetch,
      autoLogoutOnSessionExpired: autoLogoutOnSessionExpired ?? this.autoLogoutOnSessionExpired,
      maxAge: maxAge ?? this.maxAge,
      customRefreshTokenResponseParser: customRefreshTokenResponseParser ?? this.customRefreshTokenResponseParser,
      onRefreshToken: onRefreshToken ?? this.onRefreshToken,
      refreshTokenMethod: refreshTokenMethod ?? this.refreshTokenMethod,
      customRefreshTokenRequestBodyMapper:
          customRefreshTokenRequestBodyMapper ?? this.customRefreshTokenRequestBodyMapper,
      refreshTokenTimeout: refreshTokenTimeout ?? this.refreshTokenTimeout,
      onRefreshTokenFailure: onRefreshTokenFailure ?? this.onRefreshTokenFailure,
      deleteAccountApiEndpoint: deleteAccountApiEndpoint ?? this.deleteAccountApiEndpoint,
      deleteAccountCustomResponseParser: deleteAccountCustomResponseParser ?? this.deleteAccountCustomResponseParser,
      logoutAfterDelete: logoutAfterDelete ?? this.logoutAfterDelete,
    );
  }
}

/// Minimum configuration , useful for testing (fake)
BasicAuthenticationConfig basicAuthenticationConfigModel() {
  return BasicAuthenticationConfig(
    loginCredentialsAPIendpoint: () {
      return Uri.parse("example.com/login");
    },
    signupCredentialsAPIendpoint: () {
      return Uri.parse("example.com/signup");
    },
    deleteAccountApiEndpoint: (userId) {
      return Uri.parse("example.com/signup");
    },
  );
}
