import 'dart:async';
import 'dart:convert';

import 'package:corsac_jwt/corsac_jwt.dart';
import 'package:dartz/dartz.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_gb_stack_base/flutter_gb_stack_base.dart';
import 'package:logger/logger.dart';
import 'package:mockito/mockito.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:http/http.dart';

import '../../flutter_gb_authentication_basic.dart';

StreamController<Either<RefreshTokenFailure, RefreshTokenSuccess>> _authRefreshEventsController = StreamController();

Stream<Either<RefreshTokenFailure, RefreshTokenSuccess>> authRefreshEvents =
    _authRefreshEventsController.stream.asBroadcastStream();

class AuthenticationServiceImpl extends RepoServiceBaseImpl implements AuthenticationService {
  final IHttpClient httpClient;
  final BasicAuthenticationConfig config;
  final SharedPreferences sharedPreferences;

  AuthenticationServiceImpl({
    required this.httpClient,
    required this.config,
    required this.sharedPreferences,
    Logger? logger,
  }) : super(
          logger: logger ?? getAuthenticationBasicLogger(config),
          client: httpClient,
        );

  @override
  Future<Either<Failure, AuthenticationData>> loginWithCredentials(String username, String password) async {
    AuthenticationData authData;

    return wrapAndHandleHttpBaseRequest(
      () async {
        final url = config.loginCredentialsAPIendpoint();
        final requestBody = jsonEncode(
          config.customLoginRequestMapper != null
              ? (await config.customLoginRequestMapper!(username, password))
              : {
                  "email": username,
                  "password": password,
                },
        );
        logger.v("Login with credentials: $username|$password at API: $url");

        final request = Request(
          "POST",
          url,
        )..body = requestBody;
        return request;
      },
      onResponse: (response, left, right) async {
        if (config.customLoginResponseParser != null) {
          authData = await config.customLoginResponseParser!(response.body);
        } else {
          authData = defaultAuthenticationDataParser(response.body)!;
        }

        await storeAuthDataSession(authData);

        logger.v("Login successfully");
        return right(
          authData,
        );
      },
    );
  }

  @override
  Future<Either<Failure, AuthenticationData?>> signupWithCredentials(AuthSignUpPayload payload) async {
    return wrapAndHandleHttpBaseRequest(
      () {
        final url = config.signupCredentialsAPIendpoint();
        final requestBody = jsonEncode(payload.toMap());
        logger.v("Signup with credentials at API: $url");
        // logger.v("Request body is $requestBody");

        final request = Request(
          "POST",
          url,
        )..body = requestBody;
        return request;
      },
      onResponse: (response, left, right) async {
        AuthenticationData? authData;

        if (config.customSignupCredentialsResponseParser != null) {
          authData = await config.customSignupCredentialsResponseParser!(response.body);
        } else {
          authData = defaultAuthenticationDataParser(response.body)!;
        }

        /// check if we have an authData extracted from response
        /// so we can store the session
        if (authData != null && config.authenticateOnSignup) {
          await storeAuthDataSession(authData);
        }

        logger.v("Login successfully");
        return right(
          authData,
        );
      },
    );
  }

  @override
  Future<Either<Failure, AuthenticationData?>> getAuthenticatedSession() async {
    String? token;
    String? id;
    User? user;

    try {
      token = sharedPreferences.getString(AuthSharedPrefKeys.TOKEN_KEY);
      final refreshToken = sharedPreferences.getString(AuthSharedPrefKeys.REFRESH_KEY);
      final userData = sharedPreferences.getString(AuthSharedPrefKeys.USER_KEY);
      id = sharedPreferences.getString(AuthSharedPrefKeys.USER_ID_KEY);
      user = userData != null
          ? await (config.customUserMapper?.call(jsonDecode(userData)) ?? _defaultUserParser(jsonDecode(userData)))
          : null;
      if (token == null || id == null) {
        return right<Failure, AuthenticationData?>(null);
      }
      return right<Failure, AuthenticationData?>(
        AuthenticationDataModel(
          id: id,
          token: token,
          refreshToken: refreshToken,
          user: user,
        ),
      );
    } catch (e, s) {
      if (id != null && token != null) {
        return left(
          NoParsableUserFailure(
            partialData: AuthenticationDataModel(
              id: id,
              token: token,
            ),
          ),
        );
      }
      logger.e(e, e, s);
      return left(
        UnknownFailure.fromException(e),
      );
    }
  }

  @override
  Future<Either<Failure, Unit>> logout() async {
    try {
      sharedPreferences.remove(AuthSharedPrefKeys.TOKEN_KEY);
      sharedPreferences.remove(AuthSharedPrefKeys.REFRESH_KEY);
      sharedPreferences.remove(AuthSharedPrefKeys.USER_KEY);
      sharedPreferences.remove(AuthSharedPrefKeys.USER_ID_KEY);
      logger.v("Session removed, logged out");
      return right<Failure, Unit>(unit);
    } catch (e, s) {
      logger.e("Failing on logout, error: $e", e, s);
      return left<Failure, Unit>(
        UnknownFailure.fromException(e),
      );
    }
  }

  @override
  Future<Either<Failure, User>> fetchAuthUserData() async {
    return wrapAndHandleHttpBaseRequest(
      () async {
        if (config.fetchUserInformationAPIendpoint == null) {
          throw Exception("fetchUserInformationAPIendpoint configuration not provided");
        }
        final authDataRes = await getAuthenticatedSession();
        final authData = authDataRes.getOrElse(() => null);
        if (authData == null) {
          throw Exception("Cannot fetch user information since there is no session stored");
        }
        final url = config.fetchUserInformationAPIendpoint!(authData);

        return Request("GET", url);
      },
      onResponse: (response, left, right) async {
        final Map<String, dynamic> decoded = jsonDecode(response.body);
        final Map<String, dynamic> userData = Map<String, dynamic>.of(
          config.customFetchUserInformationResponseMapper?.call(
                Map<String, dynamic>.of(decoded),
              ) ??
              decoded["data"],
        );
        final user = await (config.customUserMapper?.call(userData) ?? _defaultUserParser(userData));
        return right(user!);
      },
    );
  }

  @protected
  Future<void> storeAuthDataSession(AuthenticationData authData) async {
    await sharedPreferences.setString(AuthSharedPrefKeys.TOKEN_KEY, authData.token);
    await sharedPreferences.setString(AuthSharedPrefKeys.USER_ID_KEY, authData.id);
    if (authData.refreshToken != null) {
      await sharedPreferences.setString(AuthSharedPrefKeys.REFRESH_KEY, authData.refreshToken!);
    }
    if (authData.user != null) {
      await sharedPreferences.setString(
        AuthSharedPrefKeys.USER_KEY,
        authData.user!.toJson(),
      );
    }
  }

  @override
  Future<Either<Failure, User?>> deleteAccount(String userId) async {
    return wrapAndHandleHttpBaseRequest<User?>(
      () {
        final Uri uri = config.deleteAccountApiEndpoint(userId);

        return Request("DELETE", uri);
      },
      onResponse: (response, left, right) {
        final user = config.deleteAccountCustomResponseParser?.call(
          jsonDecode(response.body),
        );
        return right(user);
      },
    );
  }

  /// If [required] is set to true, this will throw an exception if [AuthenticationDataModel] cannot be
  /// extracted from given data
  static AuthenticationDataModel? defaultAuthenticationDataParser(
    String encodedData, {
    bool required = true,
  }) {
    AuthenticationDataModel? authData;
    Map<String, dynamic> data = jsonDecode(encodedData);

    if (data.containsKey("data") && !data.containsKey("token")) {
      data = data["data"];
    }
    final String? token = data["token"];
    final String? refreshToken = data["refreshToken"];
    final user = data["user"];

    String? id = data["id"];

    if (id == null && required) {
      // debugPrint("$id");
      if (token == null) {
        throw Exception("Could not retrieve 'token' from response, response was: $encodedData");
      }
      // if null let's try to get it from token
      final splited = token.split(".");
      final payload = splited[1];

      final metadataJson = base64Decode(payload).toString();
      final data = jsonDecode(metadataJson);
      final id = data["id"];
      if (id == null) {
        throw Exception("Could not retrieve user 'id' from response, response was: $encodedData");
      }
    } else if ((id is double || id is int) && required) {
      id = id.toString();
    } else if (id is! String && required) {
      throw Exception("Invalid response, could not parse AuthenticationData, response was: $encodedData");
    }

    if (id != null && token != null) {
      authData = AuthenticationDataModel(
        id: id,
        token: token,
        refreshToken: refreshToken,
        user: user,
      );
    } else if (required) {
      throw Exception("Invalid response, could not parse AuthenticationData, response was: $encodedData");
    }

    return authData;
  }

  static defaultClient({
    Duration? maxAge,
    void Function(Map<String, String> tokens)? onRefreshToken,
    String? refreshTokenMethod,
    String Function(String token, JWT decodedToken)? refreshTokenUrl,
    required SharedPreferences sharedPreferences,
    FutureOr<Map<String, String>> Function(String refreshToken, String authToken)? customRefreshTokenRequestBodyMapper,
    FutureOr<Map<String, String>> Function(String body)? customRefreshTokenResponseParser,
    Duration? refreshTokenTimeout,
    void Function(String token, Object exception)? onRefreshTokenFailure,
  }) {
    return HttpClientImpl.defaultClient(
      sharedPreferences: sharedPreferences,
      maxAge: maxAge,
      onRefreshToken: (tokens) {
        _authRefreshEventsController.add(
          right(
            RefreshTokenSuccess(tokens: tokens),
          ),
        );
        onRefreshToken?.call(tokens);
      },
      refreshTokenMethod: refreshTokenMethod,
      refreshTokenUrl: refreshTokenUrl,
      customRefreshTokenRequestBodyMapper: customRefreshTokenRequestBodyMapper,
      customRefreshTokenResponseParser: customRefreshTokenResponseParser,
      refreshTokenTimeout: refreshTokenTimeout,
      onRefreshTokenFailure: (token, exception) {
        _authRefreshEventsController.add(
          left(
            RefreshTokenFailure(token: token, exception: exception),
          ),
        );
        onRefreshTokenFailure?.call(token, exception);
      },
    );
  }
}

// bool _sessionExpiredDeterminationClb(Map<String, dynamic> response) {
//   String message = response["message"];
//   final hasExpired = RegExp(r'expired', caseSensitive: false).hasMatch(message);
//   if (hasExpired) {
//     return true;
//   }
//   return false;
// }

FutureOr<UserModel>? _defaultUserParser(Map<String, dynamic> user) {
  return UserModel.fromMap(
    user,
  );
}

class AuthenticationServiceMock extends Mock implements AuthenticationService {
  @override
  loginWithCredentials(String? username, String? password) {
    return super.noSuchMethod(
      Invocation.method(
        #loginWithCredentials,
        [username, password],
      ),
      returnValue: Future<Either<Failure, AuthenticationData>>.value(
        right<Failure, AuthenticationData>(
          AuthenticationDataModel(token: "token", id: "id"),
        ),
      ),
      returnValueForMissingStub: Future<Either<Failure, AuthenticationData>>.value(
        right<Failure, AuthenticationData>(
          AuthenticationDataModel(token: "token", id: "id"),
        ),
      ),
    );
  }

  @override
  signupWithCredentials(AuthSignUpPayload? payload) {
    return super.noSuchMethod(
      Invocation.method(#signupWithCredentials, [payload]),
      returnValue: Future.value(right<Failure, AuthenticationData?>(null)),
      returnValueForMissingStub: Future.value(right<Failure, AuthenticationData?>(null)),
    );
  }

  @override
  getAuthenticatedSession() {
    return super.noSuchMethod(
      Invocation.method(#getAuthenticatedSession, []),
      returnValue: Future.value(right<Failure, AuthenticationData?>(null)),
      returnValueForMissingStub: Future.value(right<Failure, AuthenticationData?>(null)),
    );
  }

  @override
  Future<Either<Failure, Unit>> logout() async {
    return super.noSuchMethod(
      Invocation.method(#logout, []),
      returnValue: Future.value(right<Failure, Unit>(unit)),
      returnValueForMissingStub: Future.value(right<Failure, Unit>(unit)),
    );
  }

  @override
  Future<Either<Failure, User>> fetchAuthUserData() {
    return super.noSuchMethod(
      Invocation.method(#fetchAuthUserData, []),
      returnValue: Future.value(
        right<Failure, User>(
          userMock(),
        ),
      ),
      returnValueForMissingStub: Future.value(
        right<Failure, User>(
          userMock(),
        ),
      ),
    );
  }

  stubLoginWithCredentials([Either<Failure, AuthenticationData>? response]) {
    when(loginWithCredentials(any, any)).thenAnswer((realInvocation) async =>
        response ??
        right<Failure, AuthenticationData>(
          AuthenticationDataModel(
            id: "id",
            token: "token",
            refreshToken: "refreshToken",
            // user
          ),
        ));
  }

  stubGetSessionData([Either<Failure, AuthenticationData?>? response]) {
    when(getAuthenticatedSession()).thenAnswer((_) async {
      return response ??
          right<Failure, AuthenticationData?>(
            authenticationDataModel(),
          );
    });
  }

  stubLogoutSuccess([Either<Failure, Unit>? response]) {
    when(logout()).thenAnswer(
      (realInvocation) async => response ?? right<Failure, Unit>(unit),
    );
  }

  stbFetchUser(Either<Failure, User>? response) {
    when(fetchAuthUserData()).thenAnswer((realInvocation) async {
      return response ??
          right<Failure, User>(
            userMock(),
          );
    });
  }

  stubFailure() {
    when(loginWithCredentials(any, any)).thenAnswer((realInvocation) async => left<Failure, AuthenticationData>(
          UnknownFailure(),
        ));
  }
}

UserModel userMock() {
  return UserModel(
    (b) => b
      ..id = "id"
      ..firstName = "firstName"
      ..lastName = "lastName"
      ..username = "username"
      ..email = "test@test.com"
      ..revoked = false
      ..active = true
      ..createdAt = DateTime.now(),
  );
}
