import 'package:bloc_test/bloc_test.dart';
import 'package:flutter_gb_authentication_basic/flutter_gb_authentication_basic.dart';
import 'package:flutter_gb_stack_base/flutter_gb_stack_base.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:get_it/get_it.dart';
import 'package:mockito/mockito.dart';

final _getIt = GetIt.instance;

const tokenMock =
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyLCJleHAiOjE1MTYyMzkwMjJ9.4Adcj3UFYzPUVaVF43FmMab6RlaQD8A9V8wFzzht-KQ";

void main() {
  BasicAuthenticationConfig config = BasicAuthenticationConfig(
    loginCredentialsAPIendpoint: () {
      return Uri.parse("test");
    },
    signupCredentialsAPIendpoint: () {
      return Uri.parse("test");
    },
    deleteAccountApiEndpoint: (userId) {
      return Uri.parse("test");
    },
  );

  late AuthenticationServiceMock authenticationServiceMock;

  setUpAll(() async {
    await configureAuthenticationBasicInjection(AppEnvironment.test, config);
    authenticationServiceMock = _getIt<AuthenticationService>() as AuthenticationServiceMock;
  });

  group("loginWithCredentials interactions", () {
    blocTest<AuthenticationBasicBloc, AuthenticationBasicState>(
      "when success",
      setUp: () {
        authenticationServiceMock.stubLoginWithCredentials();
      },
      build: () => AuthenticationBasicBloc(
        config: _getIt(),
        authenticationService: authenticationServiceMock,
      ),
      act: (bloc) {
        bloc.loginWithCredentials("email", "password");
      },
      expect: () {
        return [
          const TypeMatcher<AuthenticationBasicState>()
              // .having((state) => state.authenticated, "authenticated", false)
              .having((state) => state.authenticating, "authenticating", true)
              .having((state) => state.authenticated, "authenticated", false),
          const TypeMatcher<AuthenticationBasicState>()
              // .having((state) => state.authenticated, "authenticated", false)
              .having((state) => state.authenticating, "authenticating", false)
              .having((state) => state.authenticated, "authenticated", true),
        ];
      },
    );

    blocTest<AuthenticationBasicBloc, AuthenticationBasicState>(
      "when failure",
      setUp: () {
        authenticationServiceMock.stubFailure();
      },
      build: () => AuthenticationBasicBloc(
        config: _getIt(),
        authenticationService: authenticationServiceMock,
      ),
      act: (bloc) {
        bloc.loginWithCredentials("email", "password");
      },
      expect: () {
        return [
          const TypeMatcher<AuthenticationBasicState>()
              // .having((state) => state.authenticated, "authenticated", false)
              .having((state) => state.authenticating, "authenticating", true)
              .having((state) => state.authenticated, "authenticated", false),
          const TypeMatcher<AuthenticationBasicState>()
              // .having((state) => state.authenticated, "authenticated", false)
              .having((state) => state.authenticating, "authenticating", false)
              .having((state) => state.authenticatingFailure, "authenticatingFailure", isNotNull)
              .having((state) => state.authenticated, "authenticated", false),
        ];
      },
    );
  });

  group("logout", () {
    test("service calls", () {
      final bloc = AuthenticationBasicBloc(
        authenticationService: authenticationServiceMock,
        config: _getIt(),
      );

      bloc.emit(
        bloc.state.copyWith(
          authenticationData: AuthenticationDataModel(
            id: "id",
            token: tokenMock,
          ),
        ),
      );

      bloc.logout();

      delayed(
        () {
          verify(authenticationServiceMock.logout());
        },
        duration: const Duration(milliseconds: 400),
      );
    });

    blocTest<AuthenticationBasicBloc, AuthenticationBasicState>(
      "logout",
      build: () {
        return AuthenticationBasicBloc(
          authenticationService: authenticationServiceMock,
          config: _getIt(),
        );
      },
      seed: () {
        return AuthenticationBasicState.initial().copyWith(
          authenticationData: AuthenticationDataModel(
            id: "id",
            token: tokenMock,
          ),
        );
      },
      act: (bloc) {
        bloc.logout();
      },
      wait: const Duration(milliseconds: 500),
      expect: () => [
        const TypeMatcher<AuthenticationBasicState>().having((state) => state.authenticated, "authenticated", false),
      ],
    );
  });

  test("events replay prop", () {
    final bloc = AuthenticationBasicBloc(
      authenticationService: authenticationServiceMock,
      config: _getIt(),
    );

    bloc.add(
      const AuthenticationBasicEvent.authChanged(authenticated: true, data: null),
    );
    bloc.add(
      const AuthenticationBasicEvent.logout(
        sessionToken: "sessionToken",
      ),
    );
    bloc.add(
      const AuthenticationBasicEvent.logout(
        sessionToken: "sessionToken",
      ),
    );

    expect(
        bloc.eventsReplay,
        emitsInOrder([
          const TypeMatcher<AuthChanged>(),
          const TypeMatcher<Logout>(),
          const TypeMatcher<Logout>(),
        ]));
  });
}
