import 'dart:convert';

import 'package:dartz/dartz.dart';
import 'package:flutter_gb_authentication_basic/flutter_gb_authentication_basic.dart';
import 'package:flutter_gb_stack_base/flutter_gb_stack_base.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';
import 'package:http/http.dart';

main() {
  BasicAuthenticationConfig config = BasicAuthenticationConfig(
    loginCredentialsAPIendpoint: () {
      return Uri.parse("example.com/login");
    },
    signupCredentialsAPIendpoint: () {
      return Uri.parse("example.com/signup");
    },
    deleteAccountApiEndpoint: (userId) {
      return Uri.parse("example.com/signup");
    },
  );

  late HttpClientMock clientMock;
  late SharedPreferencesMock sharedPreferencesMock;
  late AuthenticationServiceImpl authenticationService;

  setUpAll(() {
    clientMock = HttpClientMock();
    sharedPreferencesMock = SharedPreferencesMock();
    authenticationService = AuthenticationServiceImpl(
      config: config,
      httpClient: clientMock,
      sharedPreferences: sharedPreferencesMock,
    );
  });

  group("loginWithCredentials", () {
    const mockPassword = "password";
    const mockEmail = "email.com";

    group("default behavior", () {
      test("request body and response parse", () async {
        clientMock.stubSuccessSend(
          body: '{"id": "id","token":"token","refreshToken":"refreshToken","user":null}',
        );

        final res = await authenticationService.loginWithCredentials(mockEmail, mockPassword);

        verify(
          clientMock.send(
            argThat(
              // containsPair("password", mockPassword),
              const TypeMatcher<Request>().having((p0) => p0.method, "method", "POST").having(
                    (p0) => p0.body,
                    "body",
                    const TypeMatcher<String>()
                        .having((body) => jsonDecode(body)["password"], "password", mockPassword)
                        .having((body) => jsonDecode(body)["email"], "email", mockEmail),
                  ),
            ),
          ),
        );

        AuthenticationDataModel? userData = res.fold((l) => null, (r) => r as AuthenticationDataModel?);

        // expect(userData, TypeMatcher<AuthenticationDataModel>().having((p0) => p0, "password", matcher),);
        expect(userData, isNotNull);
        expect(userData!.refreshToken, "refreshToken");
        expect(userData.token, "token");
        // expect(userData.token, "token");

        verify(sharedPreferencesMock.setString(AuthSharedPrefKeys.TOKEN_KEY, "token"));
        verify(sharedPreferencesMock.setString(AuthSharedPrefKeys.REFRESH_KEY, "refreshToken"));
        verify(sharedPreferencesMock.setString(AuthSharedPrefKeys.USER_ID_KEY, "id"));
      });
    });

    group("custom behavior", () {
      test("request body map", () async {
        authenticationService = AuthenticationServiceImpl(
          config: config.copyWith(
            customLoginRequestMapper: (email, password) {
              return Map<String, String>.of({
                "username": email,
                "password": password,
              });
            },
          ),
          httpClient: clientMock,
          sharedPreferences: sharedPreferencesMock,
        );

        clientMock.stubSuccessSend(
          body: '{"id": "id","token":"token","refreshToken":"refreshToken","user":null}',
        );

        await authenticationService.loginWithCredentials(mockEmail, mockPassword);

        verify(
          clientMock.send(
            argThat(
              // containsPair("password", mockPassword),
              const TypeMatcher<Request>()
                  .having(
                    (p0) => p0.url,
                    "url",
                    config.loginCredentialsAPIendpoint(),
                  )
                  .having((p0) => p0.method, "method", "POST")
                  .having(
                    (p0) => p0.body,
                    "body",
                    const TypeMatcher<String>()
                        .having((body) => jsonDecode(body)["password"], "password", mockPassword)
                        .having((body) => jsonDecode(body)["username"], "email", mockEmail),
                  ),
            ),
          ),
        );
      });

      test("request response custom map", () async {
        authenticationService = AuthenticationServiceImpl(
          config: config.copyWith(
            customLoginResponseParser: (raw) {
              final response = jsonDecode(raw);
              return AuthenticationDataModel(
                id: response["user_id"],
                token: response["token"],
                refreshToken: response["refreshToken"],
              );
            },
          ),
          httpClient: clientMock,
          sharedPreferences: sharedPreferencesMock,
        );

        clientMock.stubSuccessSend(
          body: '{"user_id": "id","token":"token","refreshToken":"refreshToken","user":null}',
        );

        final res = await authenticationService.loginWithCredentials(mockEmail, mockPassword);

        AuthenticationDataModel? userData = res.fold((l) => null, (r) => r as AuthenticationDataModel?);

        // expect(userData, TypeMatcher<AuthenticationDataModel>().having((p0) => p0, "password", matcher),);
        expect(userData, isNotNull);
        expect(userData!.refreshToken, "refreshToken");
        expect(userData.token, "token");
        expect(userData.id, "id");
        // expect(userData.token, "token");
      });
    });
  });

  group("signupWithCredentials", () {
    // TODO: implement test
  });

  group("fetchAuthUserData", () {
    // TODO: implement test
  });

  group('deleteAccount', () {
    test('api call', () async {
      const _userId = "1";

      final res = await authenticationService.deleteAccount(_userId);

      verify(
        clientMock.send(
          argThat(const TypeMatcher<Request>()
              .having(
                (request) => request.url,
                "url",
                config.deleteAccountApiEndpoint(_userId),
              )
              .having((p0) => p0.method, "method", "DELETE")),
        ),
      );

      expect(res, isA<Right<Failure, User?>>());
    });

    test('custom response parser', () async {
      const _userId = "1";

      User? resUser;

      final user = UserModel.fromMap({
        "email": "email@email.com",
        "firstName": 'Someone',
        "createdAt": "2022-06-15 12:05:09.0",
        "id": _userId,
      });

      clientMock.stubSuccessSend(
        body: user.toJson(),
      );

      authenticationService = AuthenticationServiceImpl(
        config: config.copyWith(
          deleteAccountCustomResponseParser: (data) {
            resUser = UserModel.fromMap(data);
            return resUser;
          },
        ),
        httpClient: clientMock,
        sharedPreferences: sharedPreferencesMock,
      );

      await authenticationService.deleteAccount(_userId);

      expect(resUser, user);
    });
  });
}
